#define MUL_KARA_THRESHOLD 10
#define MUL_TOOM_THRESHOLD 41
#define NTL_GF2X_HalfGCD_CROSSOVER (MUL_KARA_THRESHOLD*NTL_BITS_PER_LONG)
#define NTL_GF2X_GCD_CROSSOVER (5*NTL_GF2X_HalfGCD_CROSSOVER)

/* #define DEBUG */

#include "mul.c"

class GF2XMatrix {
private:

   GF2XMatrix(const GF2XMatrix&);  // disable
   GF2X elts[2][2];

public:

   GF2XMatrix() { }
   ~GF2XMatrix() { }

   void operator=(const GF2XMatrix&);
   GF2X& operator() (long i, long j) { return elts[i][j]; }
   const GF2X& operator() (long i, long j) const { return elts[i][j]; }
};

/* an >= bn > 0 */
static
void mul_basecase (_ntl_ulong *cp, const _ntl_ulong *ap, long an,
            const _ntl_ulong *bp, long bn)
{
   if (an == 1) /* then bn=1 since an >= bn */
    {
      mul1 (cp, ap[0], bp[0]);
      return;
    }

  Mul1 (cp, ap, an, bp[0]);
  for (long i = 1; i < bn; i++)
    AddMul1 (cp + i, ap, an, bp[i]);
}

#if 0
void Print (const _ntl_ulong *ap, long an)
{
  cout << ap[0];
  long i;
  for (i=1;i<an;i++)
    cout << "+B^" << i << "*" << ap[i];
  cout << ":" << endl;
}
#endif

void mul_gen(GF2X& c, const GF2X& a, const GF2X& b)
{
   long sa = a.xrep.length();
   long sb = b.xrep.length();

   if (sa <= 0 || sb <= 0) {
      clear(c);
      return;
   }
 
   // finally: the general case
   
   static WordVector mem;
   static WordVector stk;
   static WordVector vec;

   const _ntl_ulong *ap, *bp;
   _ntl_ulong *cp;

   long sc = sa + sb;
   long in_mem = 0;

   if (&a == &c || &b == &c) {
      mem.SetLength(sc);
      cp = mem.elts();
      in_mem = 1;
   }
   else {
      c.xrep.SetLength(sc);
      cp = c.xrep.elts();
   }

   long n, hn, sp;

   n = min(sa, sb);

   sp = 0;
   /* the memory necessary for ToomMul is
      sp(n) = 5*ceil(n/3)+2 + sp(ceil(n/3)), and for KarMul it is
      sp(n) = 4*ceil(n/2)+3 + sp(ceil(n/2)). 
      Since 5*ceil(n/3)+2 <= 4*ceil(n/2)+3 for n > 4,
      and ceil(n/3) <= ceil(n/2), the memory allocated for KarMul is enough. */
   while (n >= MUL_KARA_THRESHOLD)
     {
       hn = (n+1) >> 1;
       sp += (hn << 2) + 3;
       n = hn;
     }

   stk.SetLength(sp);
   _ntl_ulong *stk_p = stk.elts();

   if (sa > sb) {
      { long t; t = sa; sa = sb; sb = t; }
      ap = b.xrep.elts();
      bp = a.xrep.elts();
   }
   else {
      ap = a.xrep.elts();
      bp = b.xrep.elts();
   }

   // now sa <= sb

   if (sa < MUL_KARA_THRESHOLD)
     mul_basecase (cp, bp, sb, ap, sa);
   else {

   vec.SetLength(2*sa);

   _ntl_ulong *v = vec.elts();

   long i, j;

   for (i = 0; i < sc; i++)
      cp[i] = 0;

   do {
      if (sa == 0)
        break;

      if (sa == 1)
        {
          AddMul1 (cp, bp, sb, ap[0]);
          break;
        }

      // general case

      for (i = 0; i+sa <= sb; i += sa) {
        // KarMul(v, ap, bp + i, sa, stk_p);
        ToomMul(v, ap, bp + i, sa, stk_p);
         for (j = 0; j < 2*sa; j++)
            cp[i+j] ^= v[j];
      }

      { const _ntl_ulong *t; t = ap; ap = bp + i; bp = t; }
      { long t; t = sa; sa = sb - i; sb = t; }
      cp = cp + i;
   } while (1);
   }

   if (in_mem)
      c.xrep = mem;

   c.normalize();
}

#define MUL mul_gen

void mul(GF2X& U, GF2X& V, const GF2XMatrix& M)
// (U, V)^T = M*(U, V)^T
{
   GF2X RU, RV, R1, R2;
   // RU = U
   // RV = V

   // R1 = M(0,0)
   MUL(R1, M(0,0), U);
   // R2 = M(0,1)
   MUL(R2, M(0,1), V);
   add(R2, R1, R2);

   // R1 = M(1,0)
   MUL(R1, M(1,0), U);
   U = R2; // previous value 
   // R2 = M(1,1)
   MUL(R2, M(1,1), V);
   add(V, R1, R2);
}

void mul(GF2XMatrix& A, GF2XMatrix& B, GF2XMatrix& C)
// A = B*C, B and C are destroyed
{
   long db = deg(B(1,1));
   long dc = deg(C(1,1));
   long da = db + dc;

   long k = NextPowerOfTwo(da+1);

   GF2X B00, B01, B10, B11, C0, C1, T1, T2;
   
   // C0 = C(0,0)
   // C1 = C(1,0)

   MUL(T1, B(0,0), C(0,0));
   MUL(T2, B(0,1), C(1,0));
   add(A(0,0), T1, T2);

   MUL(T1, B(1,0), C(0,0));
   MUL(T2, B(1,1), C(1,0));
   add(A(1,0), T1, T2);

   C(0,0).kill();
   C(1,0).kill();

   // C0 = C(0,1), k);  C(0,1).kill();
   // C1 = C(1,1), k);  C(1,1).kill();

   MUL(T1, B(0,0), C(0,1));
   MUL(T2, B(0,1), C(1,1));
   add(A(0,1), T1, T2);

   MUL(T1, B(1,0), C(0,1));
   MUL(T2, B(1,1), C(1,1));
   add(A(1,1), T1, T2);

   C(0,1).kill();
   C(1,1).kill();

   B(0,0).kill();
   B(1,0).kill();
   B(0,1).kill();
   B(1,1).kill();
}

void IterHalfGCD(GF2XMatrix& M_out, GF2X& U, GF2X& V, long d_red)
{
   M_out(0,0).SetMaxLength(d_red);
   M_out(0,1).SetMaxLength(d_red);
   M_out(1,0).SetMaxLength(d_red);
   M_out(1,1).SetMaxLength(d_red);

   set(M_out(0,0));   clear(M_out(0,1));
   clear(M_out(1,0)); set(M_out(1,1));

   long goal = deg(U) - d_red;

   if (deg(V) <= goal)
      return;

   GF2X Q, t(INIT_SIZE, d_red);

   while (deg(V) > goal) {
      PlainDivRem(Q, U, U, V);
      swap(U, V);

      MUL(t, Q, M_out(1,0));
      add(t, M_out(0,0), t); // add=sub over GF2
      M_out(0,0) = M_out(1,0);
      M_out(1,0) = t;

      MUL(t, Q, M_out(1,1));
      add(t, M_out(0,1), t); // add=sub over GF2
      M_out(0,1) = M_out(1,1);
      M_out(1,1) = t;
   }
}

void GF2XMatrix::operator=(const GF2XMatrix& M)
{
   elts[0][0] = M.elts[0][0];
   elts[0][1] = M.elts[0][1];
   elts[1][0] = M.elts[1][0];
   elts[1][1] = M.elts[1][1];
}

// deg(U) > deg(V),   1 <= d_red <= deg(U)+1.
//
// This computes a 2 x 2 polynomial matrix M_out such that
//    M_out * (U, V)^T = (U', V')^T,
// where U', V' are consecutive polynomials in the Euclidean remainder
// sequence of U, V, and V' is the polynomial of highest degree
// satisfying deg(V') <= deg(U) - d_red.
void HalfGCD(GF2XMatrix& M_out, const GF2X& U, const GF2X& V, long d_red, int extended)
{
   if (IsZero(V) || deg(V) <= deg(U) - d_red) {
      set(M_out(0,0));   clear(M_out(0,1));
      clear(M_out(1,0)); set(M_out(1,1));
 
      return;
   }


   long n = deg(U) - 2*d_red + 2;
   if (n < 0) n = 0;

   GF2X U1, V1;

   RightShift(U1, U, n);
   RightShift(V1, V, n);

   if (d_red <= NTL_GF2X_HalfGCD_CROSSOVER) {
      IterHalfGCD(M_out, U1, V1, d_red);
      return;
   }

   long d1 = (d_red + 1)/2;
   if (d1 < 1) d1 = 1;
   if (d1 >= d_red) d1 = d_red - 1;

   GF2XMatrix M1;

   HalfGCD(M1, U1, V1, d1, extended || (n != 0));
   mul(U1, V1, M1);

   long d2 = deg(V1) - deg(U) + n + d_red;

   if (IsZero(V1) || d2 <= 0) {
      M_out = M1;
      return;
   }


   GF2X Q;
   GF2XMatrix M2;

   DivRem(Q, U1, U1, V1);
   swap(U1, V1);

   HalfGCD(M2, U1, V1, d2, extended || (n != 0));

   GF2X t(INIT_SIZE, deg(M1(1,1))+deg(Q)+1);

   MUL(t, Q, M1(1,0));
   add(t, M1(0,0), t); // add=sub over GF2
   swap(M1(0,0), M1(1,0));
   swap(M1(1,0), t);

   t.kill();

   t.SetMaxLength(deg(M1(1,1))+deg(Q)+1);

   MUL(t, Q, M1(1,1));
   add(t, M1(0,1), t); // add=sub over GF2
   swap(M1(0,1), M1(1,1));
   swap(M1(1,1), t);

   t.kill();

   mul(M_out, M2, M1); 
}

// c <- a mod x^n, with n >= 0
void
RightShiftRem(GF2X& c, const GF2X& a, long n)
{
  if (deg(a) < n) // covers a=0 too
    {
      c = a;
      return;
    }

  long sa = a.xrep.length(); // number of words of a
  long wn = n / NTL_BITS_PER_LONG;
  long bn = n - wn * NTL_BITS_PER_LONG; // #bits of ap[wn]
  if (wn >= sa)
    {
      wn = sa;
      bn = 0;
    }

  c.xrep.SetLength (wn + (bn != 0));

   _ntl_ulong *cp = c.xrep.elts();
   const _ntl_ulong *ap = a.xrep.elts();

   long i;

   for (i = 0; i < wn; i++)
     cp[i] = ap[i];
   if (bn)
     cp[wn] = ap[wn] & ((1UL << bn) - 1UL);

   c.normalize();
}

// c <- c mod x^n, with n >= 0
void
RightShiftRem(GF2X& c, long n)
{
  if (deg(c) < n) // covers c=0 too
    return;

  long sc = c.xrep.length(); // number of words of c
  long wn = n / NTL_BITS_PER_LONG;
  long bn = n - wn * NTL_BITS_PER_LONG; // #bits of cp[wn]
  if (wn >= sc)
    {
      wn = sc;
      bn = 0;
    }

  c.xrep.SetLength (wn + (bn != 0));

   _ntl_ulong *cp = c.xrep.elts();

   if (bn)
     cp[wn] = cp[wn] & ((1UL << bn) - 1UL);

   c.normalize();
}

#if 0
int check_HalfGCD2(GF2XMatrix& M_out, GF2X& U, GF2X& V, long d_red,
		    int extended, GF2X& U0, GF2X& V0, int pt)
{
  if (!(deg(U) > deg(U0) - d_red && deg(U0) - d_red >= deg(V)))
    {
      cerr << "HalfGCD2 failed at " << pt << endl;
      cerr << "deg(U0)-d_red=" << deg(U0) - d_red << " deg(U)=" << deg(U) << " deg(V)=" << deg(V) << endl;
      return 1;
    }
  if (extended)
    {
      GF2X A, B;
      A = U0;
      B = V0;
      mul(A, B, M_out);
      add(A, A, U);
      add(B, B, V);
      if (!IsZero(A) || !IsZero(B))
	{
	  cerr << "HalfGCD2 failed at " << pt << endl;
	  A = U0;
	  B = V0;
	  mul(A, B, M_out);
	  cerr << deg(M_out(0,0)) << " " << deg(M_out(0,1)) << " " << deg(M_out(1,0)) << " " << deg(M_out(1,1)) << endl;
	  if (A != U)
	    cerr << "deg(U)=" << deg(U) << " deg(A)=" << deg(A) << endl;
	  if (V != B)
	    cerr << "deg(V)=" << deg(V) << " deg(B)=" << deg(B) << endl;
	  return 1;
	}
    }
  return 0;
}
#endif

// same as HalfGCD, except U and V are replaced by U' and V'
// if extended=0, the matrix M_out is not computed
void HalfGCD2(GF2XMatrix& M_out, GF2X& U, GF2X& V, long d_red, int extended)
{
  long degU = deg(U);

   if (IsZero(V) || deg(V) <= degU - d_red)
     {
       if (extended != 0)
         {
           set(M_out(0,0));   clear(M_out(0,1));
           clear(M_out(1,0)); set(M_out(1,1));
         }
       return;
     }

   if (d_red <= NTL_GF2X_HalfGCD_CROSSOVER)
     {
       IterHalfGCD(M_out, U, V, d_red);
       return;
     }

   long d1 = (d_red + 1)/2;
   if (d1 < 1) d1 = 1;
   if (d1 >= d_red) d1 = d_red - 1;

   long n = degU - 2 * d1 + 2;
   if (n < 0) n = 0;

   GF2X U1, V1;

   if (n != 0)
     {
       RightShiftRem (U1, U, n);
       RightShiftRem (V1, V, n);
       RightShift(U, U, n);
       RightShift(V, V, n);
     }

   GF2XMatrix M1;

   HalfGCD2(M1, U, V, d1, extended || (n != 0));

   if (n != 0)
     {
       LeftShift(U, U, n);
       LeftShift(V, V, n);
       mul(U1, V1, M1);
       add(U, U, U1);
       add(V, V, V1);
     }

   long d2 = deg(V) - (degU - d_red);

   if (IsZero(V) || d2 <= 0)
     {
       if (extended != 0)
         M_out = M1;
       return;
     }

   GF2X Q;
   GF2XMatrix M2;

   DivRem(Q, U, U, V);
   swap(U, V);

   if (extended)
     {
       GF2X t;

       MUL(t, Q, M1(1,0));
       add(t, M1(0,0), t);
       swap(M1(0,0), M1(1,0));
       swap(M1(1,0), t);

       MUL(t, Q, M1(1,1));
       add(t, M1(0,1), t);
       swap(M1(0,1), M1(1,1));
       swap(M1(1,1), t);
     }

   if (IsZero(V) || deg(V) <= degU - d_red)
     {
       if (extended)
         M_out = M1;
       return;
     }

   n = deg(U) - 2 * d2 + 2;
   if (n < 0) n = 0;

   if (n != 0)
     {
       RightShiftRem(U1, U, n);
       RightShiftRem(V1, V, n);
       RightShift(U, U, n);
       RightShift(V, V, n);
     }

   HalfGCD2(M2, U, V, d2, extended || (n != 0));

   if (n != 0)
     {
       LeftShift(U, U, n);
       LeftShift(V, V, n);
       mul(U1, V1, M2);
       add(U, U, U1);
       add(V, V, V1);
     }

   if (extended)
     mul(M_out, M2, M1);
}

void HalfGCD(GF2X& U, GF2X& V)
{
  long d_red = (2*deg(U))/3;

   if (IsZero(V) || deg(V) <= deg(U) - d_red) {
      return;
   }

   long du = deg(U);

   long d1 = (d_red + 1)/2;
   if (d1 < 1) d1 = 1;
   if (d1 >= d_red) d1 = d_red - 1;

   GF2XMatrix M1;

#define HALFGCD2
#ifndef HALFGCD2
   HalfGCD(M1, U, V, d1, 0);
   mul(U, V, M1);
#else
   HalfGCD2(M1, U, V, d1, 0);
#endif

   long d2 = deg(V) - du + d_red;

   if (IsZero(V) || d2 <= 0) {
      return;
   }

#ifndef HALFGCD2
   M1(0,0).kill();
   M1(0,1).kill();
   M1(1,0).kill();
   M1(1,1).kill();
#endif

   GF2X Q;

   DivRem(Q, U, U, V);
   swap(U, V);

#ifndef HALFGCD2
   HalfGCD(M1, U, V, d2, 0);
   mul(U, V, M1);
#else
   HalfGCD2(M1, U, V, d2, 0);
#endif
}


void FastGCD(GF2X& d, const GF2X& u, const GF2X& v)
{
   GF2X u1, v1;

   u1 = u;
   v1 = v;

   if (deg(u1) == deg(v1)) {
      if (IsZero(u1)) {
         clear(d);
         return;
      }

      rem(v1, v1, u1);
   }
   else if (deg(u1) < deg(v1)) {
      swap(u1, v1);
   }

   // deg(u1) > deg(v1)

   while (deg(u1) > NTL_GF2X_GCD_CROSSOVER && !IsZero(v1)) {
     HalfGCD(u1, v1);

      if (!IsZero(v1)) {
         rem(u1, u1, v1);
         swap(u1, v1);
      }
   }

   GCD(d, u1, v1);
}

