/* check-ntl - NTL program to check log files produced by the 'irred' program.

  Copyright 2003, 2004 Paul Zimmermann.
  Version 1.7, 19 January 2003.

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the
  Free Software Foundation; either version 2 of the License, or (at your
  option) any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; see the file COPYING.  If not, write to the Free
  Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
  02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <NTL/GF2XFactoring.h>

NTL_CLIENT

/* checks an input file with lines of the form

         r s k
   or    r s xnnnn
   or    r s ynnnn
   or    r s primitive

Example:

% cat in
1279 1 4
1279 2 x0b6d4db0
1279 2 yb7daa158
1279 216 primitive
3021377 135 xdcea0ee
3021377 1544 y8a270db4

% ./check-ntl < in
This is check-ntl version 1.7
1279 1 4 NTL 
1279 2 x0b6d4db0 NTL
1279 2 yb7daa158 NTL
1279 216 primitive NTL 
3021377 135 xdcea0ee NTL 
3021377 1544 y8a270db4 NTL
0 error(s)

*/

#include "HalfGCD.c"

int
main (int argc, char *argv[])
{
  int r, s, i, k, errors = 0;
  char c0, c;
  unsigned long low_bits;
  int verbose = 0;
  int fast = 0;

  fprintf (stderr, "This is check-ntl version 1.7\n");

  if (argc > 1 && strcmp (argv[1], "-v") == 0)
    {
      verbose = 1;
      argc --;
      argv ++;
    }

  if (argc > 1 && strcmp (argv[1], "-fast") == 0)
    {
      fast = 1;
      cout << "Using FastGCD\n";
      argc --;
      argv ++;
    }

  if (argc != 1)
    {
      fprintf (stderr, "Usage: check-ntl < file\n");
      exit (1);
    }

  while (1)
    {
      if (scanf ("%d %d", &r, &s) == EOF)
	break;

      printf ("%d %d ", r, s);

      do
	{
	  c = getchar ();
	}
      while (c == ' ');

      c0 = c; /* either p, or x or y, or a digit */
      if (c == 'p')
	{
	  scanf ("rimitive\n");
	  k = r;
	  low_bits = 2;
	}
      else if (c == 'x' || c == 'y')
	{
	  for (low_bits = 0, i=0; i<8; i++)
	    {
	      c = getchar ();
	      low_bits <<= 4;
	      if ('0' <= c && c <= '9')
		low_bits += c - '0';
	      else if ('a' <= c && c <= 'f')
		low_bits += (c - 'a') + 10;
	      else if (c != ' ')
		{
		  fprintf (stderr, "Error, unexpected character '%c'\n", c);
		  exit (1);
		}
	    }
	  k = r;
	}
      else
	{
	  ungetc (c, stdin);
	  scanf ("%u\n", &k);
	}

  GF2X p;
  SetCoeff(p, r);
  SetCoeff(p, 0);

  GF2X a, x;
  SetX(x);

  SetCoeff(p, s, 1);
  GF2XModulus F(p);
  /* first compute x^(2^k) or xr^(2^k) for y...,
     where xr = 1/x mod p = x^(r-1) + x^(s-1) */
  if (c0 == 'y')
    {
      SetCoeff(a, r - 1);
      SetCoeff(a, s - 1);
    }
  else
    a = x;
  for (i=1; i<=k; i++)
    {
      SqrMod(a, a, F);
      if (verbose && i % 1000 == 1)
        fprintf (stderr, "step %u, estimated time=%f\n",
                 i, GetTime () * (double) k / (double) i);
    }
  if (k == r) /* case x... or y... or primitive */
    {
      if (c0 == 'p') /* case primitive */
	{
	  a = a - x; 
	  if (!IsZero(a))
	    {
	      errors ++;
	      printf (" *** error ***");
	    }
	  else
	    printf ("primitive");
	  fflush (stdout);
	}
      else /* case c0 = x or y */
	{
	  unsigned long bits = 0;
	  int j;

          /* if c0=y, compute x^(r-1)*(1/x)^(2^r) */
          if (c0 == 'y')
            {
              GF2X b;
              SetCoeff(b, r - 1);
              a = a * b % p;
            }
	  for (j=31; j>=0;j--)
	      bits = (bits << 1) +
                (unsigned long) rep(coeff(a, (c0 == 'x') ? j : (r-1-j)));

	  if (bits != low_bits)
	    {
	      errors ++;
	      printf (" *** error: expected %c%x, got %c%x", c0, low_bits,
		      c0, bits);
	    }
	  else
	    printf ("%c%x", c0, low_bits);
	  fflush (stdout);
	}
    }
  else 
    {
      /* Remark: we only check here that x^r+x^s+1 has a factor of degree k,
	 and don't check that it has no factor of smaller degree */
      a = a - x; /* x^(2^k)+x */
      double st = GetTime ();
      if (fast)
        FastGCD(a, a, p);
      else
        GCD(a, a, p);
      if (verbose)
        printf ("GCD took %f\n", GetTime () - st);
      if (deg(a) == 0)
	{
	  errors ++;
	  printf (" *** error ***");
	}
      else
	printf ("%u", k);
      fflush (stdout);
    }
  printf (" NTL-1.7\n");
  fflush (stdout);
    }

  fprintf (stderr, "%u error(s)\n", errors);
  fflush (stderr);

  return errors;
}
