#include <stdio.h>
#include <stdlib.h>
#include <limits.h> /* for CHAR_BIT */
#include <string.h>
#include <NTL/GF2XFactoring.h>
#include <NTL/ZZXFactoring.h>

NTL_CLIENT

#define EXTENDED /* for extended log format */

/* usage: sieve-ntl [-v] [-fast] [-start s0] r k [k2] < file
   where file contains lines of the form

      s nnn
   or s u.

   "s nnn" means that x^r+x^s+1 has a factor of degree nnn,
   and this is output as is.

   Starts from s=s0 (default is 1).

   "s u" means that the status of x^r+x^s+1 is unknown,
   and "s k" is output if it has a factor of degree k,
   otherwise "s u" is output.
*/

#include "HalfGCD.c"

/* prints a in hexadecimal (leading bits first) */
void
print_hex (char *s, GF2X a)
{
  long d = deg (a);
  long n = (d + 4) / 4; /* number of characters */
  long i, c;
  char t[17] = "0123456789abcdef";

  c = 0;
  for (i = d; i >= 0; i--)
    {
      c = (2 * c) + rep (coeff (a, i));
      if (i % 4 == 0)
        {
          *s++ = t[c];
          c = 0;
        }
    }
  *s = '\0';
}

void FastMulMod(GF2X& c, const GF2X& a, const GF2X& b, const GF2XModulus& F)
{
   if (F.n < 0) Error("MulMod: uninitialized modulus");

   GF2X t;
   mul_gen(t, a, b);
   rem(c, t, F);
}

int
main (int argc, char *argv[])
{
  int r, k, k1, k2, s, i, n, unknown = 0, killed = 0, twokm1, s0 = 1;
  int small;
  int *saved;
  char c;
#ifdef EXTENDED
  char *fact;
#endif

  int verbose = 0;
  int fast = 0;

  fprintf (stderr, "This is sieve-ntl version 1.9\n");

  if (argc > 1 && strcmp (argv[1], "-v") == 0)
    {
      verbose = 1;
      argc --;
      argv ++;
    }

  if (argc > 1 && strcmp (argv[1], "-fast") == 0)
    {
      fast = 1;
      argc --;
      argv ++;
    }

  if (strcmp (argv[1], "-start") == 0)
    {
      s0 = atoi (argv[2]);
      argv += 2;
      argc -= 2;
    }

  if (argc != 3 && argc != 4)
    {
      fprintf (stderr, "Usage: sieve-ntl r k [k2] < file\n");
      exit (1);
    }

  r = atoi (argv[1]);
  k = atoi (argv[2]);
  k2 = (argc == 3) ? k : atoi (argv[3]);

  if (verbose)
    {
      if (fast)
        cerr << "Using FastGCD\n";
      fprintf (stderr, "Sieving for degrees %d..%d", k, k2);
      if (s0 > 1)
        fprintf (stderr, ", s0=%d", s0);
      fprintf (stderr, "\n");
    }

#ifdef EXTENDED
  fact = (char*) malloc ((2 + (k2 / 4)) * sizeof (char));
#endif

  /* since all degree k factors divide x^(2^k)-x, if x^r+x^s+1 has a degree
     k factor, then so has x^r+x^(s+2^k-1)+1, thus if 2^k-1 <= floor(r/2)-1,
     i.e. 2^(k+1) <= r, then we can save some computations */
  small = (k == k2) && (k < sizeof(int) * CHAR_BIT) && (r >> k) >= 2;
  if (small)
    {
      fprintf (stderr, "Degree is small: storing discarded degrees\n");
      twokm1 = (1 << k) - 1; /* 2^k-1 */
      saved = (int*) malloc (twokm1 * sizeof (int));
      for (i=0; i<twokm1; i++)
        saved[i] = -1; /* not computed */
    }

  while (1)
    {
      if (scanf ("%d ", &s) == EOF)
        break;

      c = getchar ();

      if (c != 'u') /* just copies rest of line */
        {
          if (s >= s0)
            printf ("%d %c", s, c);
          do
            {
              c = getchar ();
              if (s >= s0)
                putchar (c);
            }
          while (c != '\n');
        }
      else /* unknown status */
        {
          int divisible;

          scanf ("\n");

	  if (s < s0) /* skip */
	    divisible = 0;
          else if ((small == 0) || (saved[s % twokm1] == -1))
            {
              GF2X p;
              SetCoeff(p, r);
              SetCoeff(p, 0);

              GF2X a, b, c, x;
              SetX(x);

              SetCoeff(p, s, 1);
              GF2XModulus F(p);
              double st = GetTime ();
              /* check that x^(2^r) = x */
              b = x;

              for (i=0; i<k; i++)
                SqrMod(b, b, F); /* b = x^(2^k) */

              a = b - x; /* x^(2^k) - x */
              if (verbose)
                printf ("Exponentiation took %f\n", GetTime () - st);

              st = GetTime ();
              for (i = k + 1; i <= k2; i++)
                {
                  SqrMod(b, b, F); /* b = x^(2^i) */
                  c = b - x;
                  FastMulMod(a, a, c, F);
                }
              if (verbose && k < k2)
                printf ("Products took %f\n", GetTime () - st);
              
              st = GetTime ();
              if (fast)
                FastGCD(a, a, p);
              else
                GCD(a, a, p);
              if (verbose)
                printf ("GCD took %f\n", GetTime () - st);

              divisible = deg(a) != 0;
              if (divisible)
                {
                  k1 = k; // degree of found factor
                  if (deg (a) != k)
                    {
                      vec_pair_GF2X_long factors;
                      CanZass (factors, a, 0);
                      long i0 = 0;
                      // find factor of smallest degree
                      for (i = 1; i < factors.length(); i++)
                        if (deg(factors[i].a) < deg(factors[i0].a))
                          i0 = i;
                      a = factors[i0].a; // factor of smallest degree
                      k1 = deg(a);
                    }
                  print_hex (fact, a);
                }
              if (small)
                saved[s % twokm1] = divisible;
            }
          else
            divisible = saved[s % twokm1];

          if (!divisible)
            {
	      if (s >= s0)
		printf ("%d u\n", s);
              unknown ++;
            }
          else
            {
              if (s >= s0)
                {
                  printf ("%d %d", s, k1);
                  if (!small)
                    printf (" p%s", fact);
                  printf ("\n");
                }
              killed ++;
            }
        }
      fflush (stdout);
    }
  //  fprintf (stderr, "removed %u, remaining %u of unknown status\n", killed, unknown);

  if (small)
    free (saved);
#ifdef EXTENDED
  free (fact);
#endif

  return 0;
}

